/*

dump_d64 - read files from d64 files.

Copyright 2012 by Michael Kohn
http://www.mikekohn.net/
mike@mikekohn.net

Licensed under the GPL

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d64_util.h"

#if 0
static unsigned char d64_sector_count[] = {
  21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21,
  19, 19, 19, 19, 19, 19, 19, 18, 18, 18, 18, 18, 18, 17, 17, 17, 17,
  17, 17, 17, 17, 17, 17
};
#endif

static unsigned short int d64_sector_sum[] = {
  0, 21, 42, 63, 84, 105, 126, 147, 168, 189, 210, 231, 252, 273, 294,
  315, 336, 357, 376, 395, 414, 433, 452, 471, 490, 508, 526, 544, 562,
  580, 598, 615, 632, 649, 666, 683, 700, 717, 734, 751
};

static char *d64_basic_tokens[] = {
  "END", "FOR", "NEXT", "DATA", "INPUT#", "INPUT", "DIM", "READ", "LET",
  "GOTO", "RUN", "IF", "RESTORE", "GOSUB", "RETURN", "REM", "STOP", "ON",
  "WAIT", "LOAD", "SAVE", "VERIFY", "DEF", "POKE", "PRINT#", "PRINT",
  "CONT", "LIST", "CLR", "CMD", "SYS", "OPEN", "CLOSE", "GET", "NEW",
  "TAB(", "TO", "FN", "SPC(", "THEN", "NOT", "STEP", "+", "-", "*", "/",
  "^", "AND", "OR", "<", "=", ">", "SGN", "INT", "ABS", "USR", "FRE", "POS",
  "SQR", "RND", "LOG", "EXP", "COS", "SIN", "TAN", "ATN", "PEEK", "LEN",
  "STR$", "VAL", "ASC", "CHR$", "LEFT$", "RIGHT$", "MID$", "GO"
};

#if 0
static int d64_get_sector_count(int track)
{
  if (track>40) return 0;
  return d64_sector_count[track-1];
}
#endif

static int d64_get_offset(int track, int sector)
{
  sector+=d64_sector_sum[track-1];

  return sector*256;
}

static int d64_read_sector(FILE *in, int track, int sector, unsigned char *sector_data)
{
int i;

  fseek(in, d64_get_offset(track, sector), SEEK_SET);
  i=fread(sector_data, 1, 256, in);

  return i;
}

static void d64_hexdump_sector(unsigned char *sector_data)
{
unsigned char data[16];
int n,i,ptr=0;

  data[0]=0;

  for (n=0; n<256; n++)
  {
    data[ptr++]=sector_data[n];
    printf("%02x ", sector_data[n]);

    if (((n+1)%16)==0)
    {
      for (i=0; i<16; i++)
      {
        if (data[i]>=32 && data[i]<127) { printf("%c", data[i]); }
        else { printf("."); }
      }
      printf("\n");
      ptr=0;
    }
  }
}

void d64_dump_bam(FILE *in)
{
unsigned char sector_data[256];

  printf("---------------- Disk BAM -----------------------\n");
  d64_read_sector(in, 18, 0, sector_data);
  d64_hexdump_sector(sector_data);
  printf("\n");
}

void d64_hexdump_dir(FILE *in)
{
unsigned char sector_data[256];
int track=18, sector=1;

  printf("-------------- Disk DIR Hexdump ------------------\n");

  while(track>0)
  {
    d64_read_sector(in, track, sector, sector_data);
    d64_hexdump_sector(sector_data);
    printf("\n");
    track=sector_data[0];
    sector=sector_data[1];
  }
}

void d64_dir(FILE *in)
{
char *filetypes[] = { "DEL", "SEQ", "PRG", "USR", "REL" };
unsigned char sector_data[256];
int track=18, sector=1;
int offset=0;
int i;

  printf("---------------- Disk DIR -----------------------\n");

  while(track>0)
  {
    d64_read_sector(in, track, sector, sector_data);
    for (offset=0; offset<256; offset+=32)
    {
      i=sector_data[offset+2]&0x0f; 
      if (i>4 || i==0) continue;
      printf("%s ", filetypes[i]);
      for (i=0; i<16; i++)
      {
        char ch=sector_data[offset+i+5];
        if (ch>=' ' && ch<=120)
        { printf("%c", ch); }
          else
        { printf(" "); }
      }
      printf(" %d", sector_data[offset+31]<<8|sector_data[offset+30]);
      printf(" (%d,%d)\n", sector_data[offset+3], sector_data[offset+4]);
     
    }
    track=sector_data[0];
    sector=sector_data[1];
  }
}

static void d64_find_file(FILE *in, char *filename, int *track, int *sector)
{
unsigned char sector_data[256];
int ptr,i=0;
int offset;
int t=18, s=1;

  while(t>0)
  {
    d64_read_sector(in, t, s, sector_data);
    for (offset=0; offset<256; offset+=32)
    {
      ptr=0;
      for (i=0; i<16; i++)
      {
        char ch=sector_data[offset+i+5];
        //printf("%d) ch=%02x  infile=%02x\n", i, (unsigned char)ch, infile[ptr]);
        if (filename[ptr]==0)
        {
          if ((unsigned char)ch!=0xa0) { break; }
        }
          else
        {
          if (ch!=filename[ptr++]) { break; }
        }
      }

      if (i==16)
      {
        *track=sector_data[offset+3];
        *sector=sector_data[offset+4];
        //printf("track=%d sector=%d\n", t, sector);
        return;
      }
    }

    t=sector_data[0];
    s=sector_data[1];
  }

  *track = 0;
  *sector = 0;
}

int d64_dump_file(FILE *in, char *infile, char *outfile)
{
unsigned char sector_data[256];
int track, sector;
FILE *out=NULL;
int i;

  d64_find_file(in, infile, &track, &sector);

  if (track==0) return -1;

  if (outfile!=NULL)
  {
    out=fopen(outfile, "wb");
    if (out==NULL) { return -2; }
    printf("Saving to %s\n", outfile);
  }

  while(track>0)
  {
    d64_read_sector(in, track, sector, sector_data);
    if (out!=NULL)
    {
      i=fwrite(sector_data+2, 1, 254, out);
    }
      else
    {
      printf("---------- track %d / sector %d -----------\n", track, sector);
      d64_hexdump_sector(sector_data);
    }

    track=sector_data[0];
    sector=sector_data[1];
  }

  if (out!=NULL) { fclose(out); }

  return 0;
}

void d64_basic_file_list(struct _basic *basic)
{
int address;
int instring=0;
int i;

  printf("start=%d\n", basic->start);
  printf("end=%d\n", basic->end);

  printf("\n");

  for (i=0; i<65536; i++)
  {
    address=basic->line_address[i];
    if (address==0) continue;

    printf("%d ", i);

    while (basic->ram[address]!=0)
    {
      unsigned char ch=basic->ram[address++];

      if (instring==1)
      {
        if (ch>=' ' && ch<=126)
        { printf("%c", ch); }
          else
        { printf("?"); }

        if (ch=='"') { instring=0; }
      }
        else
      if (ch>=128 && ch<=203)
      {
        printf("%s", d64_basic_tokens[ch-128]);
      }
        else
      {
        if (ch=='"') { instring=1; }
        printf("%c", ch);
      }
    }

    printf("\n");
  }
}

static void d64_print_html(unsigned char *s)
{
  while(*s!=0)
  {
    printf("<img src='lowercase/%d.gif'>", *s);
    s++;
  }
}

void d64_basic_file_html(struct _basic *basic)
{
unsigned char line_str[7];
unsigned char ch[2];
int address;
int instring=0;
int i;

  printf("start=%d\n", basic->start);
  printf("end=%d\n", basic->end);

  printf("\n");

  ch[1]=0;

  printf("<html>\n");
  printf("<body>\n");
  printf("<p>\n");

  for (i=0; i<65536; i++)
  {
    address=basic->line_address[i];
    if (address==0) continue;

    sprintf((char *)line_str, "%d ", i);
    d64_print_html(line_str);

    while (basic->ram[address]!=0)
    {
      *ch=basic->ram[address++];

      if (instring==1)
      {
        d64_print_html(ch);

        if (*ch=='"') { instring=0; }
      }
        else
      if (*ch>=128 && *ch<=203)
      {
        d64_print_html((unsigned char *)d64_basic_tokens[(*ch)-128]);
      }
        else
      {
        if (*ch=='"') { instring=1; }
        d64_print_html(ch);
      }
    }

    printf("<br>\n");
  }

  printf("</p>\n");
  printf("</body>\n");
  printf("</html>\n");
}

int d64_basic_file_load(FILE *in, char *filename, struct _basic *basic)
{
unsigned char sector_data[256];
int track, sector;
int first=0;
int line=0;
int state=0;
int address=0;
int next_address=0;
int instring=0;
int i;

  d64_find_file(in, filename, &track, &sector);

  if (track==0) return -1;

  memset(basic, 0, sizeof(struct _basic));
  basic->start=65536;
  basic->end=0;

  while(track>0)
  {
    d64_read_sector(in, track, sector, sector_data);
    if (first==0)
    {
      next_address=sector_data[2]|(sector_data[3]<<8);
      first=4;
    }
    else { first=2; }

    i=first;
    while (i<256)
    {
      if (state==0)
      {
        address=sector_data[i];
        state++;
      }
        else
      if (state==1)
      {
        address|=(sector_data[i]<<8);
        if (address==0)
        {
          state=0;
        }
        state++;
      }
        else
      if (state==2)
      {
        line=sector_data[i];
        state++;
      }
        else
      if (state==3)
      {
        line|=(sector_data[i]<<8);
        basic->line_address[line]=next_address;
        next_address=address;
        address=basic->line_address[line];

        if (address<basic->start)
        {
          basic->start=address;
        }

        state++;
      }
        else
      {
        basic->ram[address++]=sector_data[i];
        if (instring==1)
        {
          if (sector_data[i]=='"') { instring=0; }
        }
          else
        if (sector_data[i]==0)
        {
          if (address>basic->end) basic->end=address;
          state=0;
        }
          else
        {
          if (sector_data[i]=='"') { instring=1; }
        }
      }

      i++;

      if (address>=65536)
      {
        printf("Address bigger than 65536.  Problem with BASIC file.\n");
        return -3;
      }
    }

    track=sector_data[0];
    sector=sector_data[1];
  }

  return 0;
}

int d64_basic_file_dump(FILE *in, char *filename)
{
unsigned char sector_data[256];
int track, sector;
int start=0;
int line=0;
int state=0;
int address=0;
int instring=0;
int i;

  d64_find_file(in, filename, &track, &sector);

  if (track==0) return -1;

  while(track>0)
  {
    d64_read_sector(in, track, sector, sector_data);
    if (start==0) { start=4;}
    else { start=2; }

    i=start;
    while (i<256)
    {
      if (state==0)
      {
        address=sector_data[i];
        state++;
      }
        else
      if (state==1)
      {
        address|=(sector_data[i]<<8);
        if (address==0) state=0;
        state++;
      }
        else
      if (state==2)
      {
        line=sector_data[i];
        state++;
      }
        else
      if (state==3)
      {
        printf("[%d] %d ", address, line|(sector_data[i]<<8));
        state++;
      }
        else
      {
        if (instring==1)
        {
          if (sector_data[i]>=' ' && sector_data[i]<=126)
          { printf("%c", sector_data[i]); }
            else
          { printf("?"); }
          if (sector_data[i]=='"') { instring=0; }
        }
          else
        if (sector_data[i]==0)
        {
          printf("\n");
          state=0;
        }
          else
        if (sector_data[i]>=128 && sector_data[i]<=203)
        {
          printf("%s", d64_basic_tokens[sector_data[i]-128]);
        }
          else
        {
          if (sector_data[i]=='"') { instring=1; }
          printf("%c", sector_data[i]);
        }
      }

      i++;
    }

    track=sector_data[0];
    sector=sector_data[1];
  }

  return 0;
}



