/*

dump_d64 - read files from d64 files.

Copyright 2012 by Michael Kohn
http://www.mikekohn.net/
mike@mikekohn.net

Licensed under the GPL

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d64_util.h"

int main(int argc, char *argv[])
{
FILE *in;
char *filename="default.d64";
char *infile=NULL;
char *outfile=NULL;
int ret;
int command=-1;
int n;

  printf("\ndump_d64 - Copyright 2012 by Michael Kohn\n");
  printf("  Webpage: http://www.mikekohn.net/\n");
  printf("  Version: February 8, 2012\n\n");

  if (argc<2)
  {
    printf("Usage: dump_d64 [ options ] <filename.d64>\n");
    printf("  -dir              Directory listing\n");
    printf("  -dir_dump         Directory hexdump\n");
    printf("  -list <filename>  Show BASIC listing of file in d64 image\n");
    printf("  -list_dump <in>   Same as -list but dumps all out of order\n");
    printf("  -list_html <in>   Same as -list but dumps all in html\n");
    printf("  -dump <in> <out>  Dump input file to out file\n");
    printf("  -dumphex <in>     Dump input file to hex output\n");
    printf("  -bam              Dump bam to screen\n");
    printf("\n");
    exit(0);
  }

  for(n=0; n<argc; n++)
  {
    if (strcmp(argv[n], "-dir")==0)
    {
      command=0;
    }
      else
    if (strcmp(argv[n], "-dir_dump")==0)
    {
      command=1;
    }
      else
    if (strcmp(argv[n], "-list")==0)
    {
      command=2;
      infile=argv[++n];
    }
      else
    if (strcmp(argv[n], "-list_dump")==0)
    {
      command=5;
      infile=argv[++n];
    }
      else
    if (strcmp(argv[n], "-list_html")==0)
    {
      command=6;
      infile=argv[++n];
    }
      else
    if (strcmp(argv[n], "-dump")==0)
    {
      command=3;
      infile=argv[++n];
      outfile=argv[++n];
    }
      else
    if (strcmp(argv[n], "-dumphex")==0)
    {
      command=3;
      infile=argv[++n];
    }
      else
    if (strcmp(argv[n], "-bam")==0)
    {
      command=4;
    }
      else
    {
      filename=argv[n];
    }
  }

  if (command==-1)
  {
    printf("Error: Unknown command.\n");
    exit(1);
  }

  printf("Opening %s\n", filename);

  in=fopen(filename, "rb");
  if (in==NULL)
  {
    printf("Error: Couldn't open file %s\n", filename);
    exit(1);
  }

  ret=0;

  if (command==0)
  {
    d64_dir(in);
  }
    else
  if (command==1)
  {
    d64_hexdump_dir(in);
  }
    else
  if (command==2)
  {
    struct _basic basic;
    ret=d64_basic_file_load(in, infile, &basic);
    if (ret==0) { d64_basic_file_list(&basic); }
  }
    else
  if (command==3)
  {
    ret=d64_dump_file(in, infile, outfile);
  }
    else
  if (command==4)
  {
    d64_dump_bam(in);
  }
    else
  if (command==5)
  {
    d64_basic_file_dump(in, infile);
  }
    else
  if (command==6)
  {
    printf("Font set can be downloaded Joe Davisson: http://joe.csoft.net/\n");
    struct _basic basic;
    ret=d64_basic_file_load(in, infile, &basic);
    if (ret==0) { d64_basic_file_html(&basic); }
  }

  if (ret==-1) printf("File Not Found\n");
    else
  if (ret==-2) printf("Error?\n");

  fclose(in);

  return 0;
}



